# -*- coding: utf-8 -*-
#
# This file is part of fedrepo_req.
# Copyright © 2017 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
"""
Provides helper functions for Bugzilla
"""
from __future__ import absolute_import
from six.moves.urllib.parse import urlencode
from datetime import datetime
from bugzilla import Bugzilla
# In version 2, BugzillaError was moved to bugzilla.transport
try:
    from bugzilla.transport import BugzillaError
except ImportError:
    from bugzilla import BugzillaError

from fedrepo_req.exceptions import ValidationError
from fedrepo_req import FAS_CLIENT


class BugzillaClient(object):
    """
    A helper class to maintain a Bugzilla session
    """
    def __init__(self):
        self.url = 'https://bugzilla.redhat.com'
        self.api_url = '{0}/xmlrpc.cgi'.format(self.url)
        self.client = Bugzilla(self.api_url)

    def login(self, username, password):
        """
        A login function to Bugzilla
        :param username: a string of the username
        :param password: a string of the password
        :return: None
        """
        try:
            self.client.login(username, password)
        except BugzillaError:
            raise ValidationError(
                'The username or password entered is invalid')

    def get_valid_review_bug(self, bug_id, pkg, require_auth=False,
                             check_fas=False):
        """
        Checks the validity of a Bugzilla bug representing a Fedora package
        review. This function was inspired by:
        https://github.com/fedora-infra/pkgdb2/blob/master/pkgdb2/api/extras.py
        https://pagure.io/pkgdb-cli/blob/master/f/pkgdb2client/utils.py
        :param bug_id: string or integer of the Bugzilla ticket ID
        :param pkg: string of the package name
        :param require_auth: a boolean determining if an exception is thrown
        if the client is not authenticated
        :param check_fas: a boolean determining if the approver is a packager
        :return: Bugzilla bug object
        """
        # When authenticated, you get a user's email address instead of their
        # full name
        if require_auth and not self.client.logged_in:
            raise ValidationError('The Bugzilla client is not authenticated')

        try:
            bug = self.client.getbug(bug_id)
        except:
            raise ValidationError('The Bugzilla ticket provided is invalid')
        # Check that the ticket is valid
        if bug.component != 'Package Review':
            raise ValidationError(
                'The Bugzilla ticket provided is not the proper type')
        elif bug.product != 'Fedora':
            raise ValidationError(
                'The Bugzilla ticket provided is not for Fedora')
        elif bug.assigned_to in ['', None, 'nobody@fedoraproject.org']:
            raise ValidationError(
                'The ticket provided is not assigned to anyone')
        # Check if the review was approved and by whom
        flag_set = False
        for flag in bug.flags:
            if flag.get('name') == 'fedora-review':
                if flag.get('status') == '+':
                    flag_set = True
                if check_fas:
                    fas_user = FAS_CLIENT.get_fas_user_by_email(flag['setter'])
                    if not FAS_CLIENT.user_member_of(fas_user, 'packager'):
                        raise ValidationError('The Bugzilla ticket\'s review '
                                              'is approved by a user that is '
                                              'not a packager')
                if flag['setter'] == bug.creator:
                    error = ('The Bugzilla ticket\'s review is approved by '
                             'the person creating the ticket. This is not '
                             'allowed.')
                    raise ValidationError(error)
                elif flag['setter'] != bug.assigned_to:
                    raise ValidationError('The review is not approved by the '
                                          'assignee of the ticket')
                # TODO: Should we still check this?
                update_dt = flag.get('modification_date')
                if update_dt:
                    dt = datetime.strptime(
                        update_dt.value, '%Y%m%dT%H:%M:%S')
                    delta = datetime.utcnow().date() - dt.date()
                    if delta.days > 60:
                        raise ValidationError('The Bugzilla ticket\'s review '
                                              'was approved over 60 days ago')
                break
        if not flag_set:
            raise ValidationError(
                'The Bugzilla ticket is not approved yet')
        # Check the format of the Bugzilla ticket title
        tmp_summary = bug.summary.partition(':')[2]
        if not tmp_summary:
            raise ValidationError(
                'Invalid title for this Bugzilla ticket (no ":" present)')
        if ' - ' not in tmp_summary:
            raise ValidationError(
                'Invalid title for this Bugzilla ticket (no "-" present)')
        pkg_in_ticket = tmp_summary.split(' - ', 1)[0].strip()
        if pkg != pkg_in_ticket:
            error = ('The package in the Bugzilla ticket "{0}" doesn\'t match '
                     'the one provided "{1}"'.format(pkg_in_ticket, pkg))
            raise ValidationError(error)
        return bug

    def get_bug_url(self, bug_id):
        """
        Forms the url to a Bugzilla ticket
        :param bug_id: a string of the bug ID
        :return: a string of the Bugzilla ticket URL
        """
        bz_query_params = urlencode({'id': bug_id})
        return '{0}/show_bug.cgi?{1}'.format(
            self.url.rstrip('/'), bz_query_params)
