# -*- coding: utf-8 -*-
#
# This file is part of fedrepo_req.
# Copyright © 2017 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
"""
Pagure API helper functions
"""
from __future__ import absolute_import
from six.moves.urllib.parse import urlencode

from fedrepo_req import CONFIG
from fedrepo_req.config import get_config_item
from fedrepo_req.request_utils import (
    get_auth_header, requests_wrapper, get_request_json)


def get_pagure_auth_header(access='app', token_type='ticket'):
    """
    Get a dictionary of the headers needed to make an authenticated API
    call to Pagure
    :param access: a string of the level of access (e.g. app or admin)
    :param token_type: a string of the type of token used (e.g. ticket or
    global)
    :return: a dictionary of the headers needed to make an authenticated API
    call to Pagure
    """
    prefix = 'pagure'
    if access == 'admin' and token_type == 'ticket':
        prefix = 'pagure_ticket'
    return get_auth_header(prefix, access)


def get_project(namespace, repo):
    """
    Get an existing Pagure project
    :param namespace: a string representing the namespace to create the project
    in
    :param repo: a string of the project/repo name
    :return: a dictionary representing the project or None
    """
    pagure_url = get_config_item(CONFIG, 'app', 'pagure_dist_git_url')
    pagure_project_url = \
        '{0}/api/0/{1}/{2}'.format(pagure_url.rstrip('/'), namespace, repo)

    rv = requests_wrapper(
        pagure_project_url, timeout=60, service_name='Pagure')
    if rv.status_code == 404:
        return None
    else:
        return get_request_json(rv, 'getting a project in Pagure')


def get_issue(issue_id):
    """
    Get an issue from the ticket queue
    :return: a dictionary representing the issue or None
    """
    pagure_url = get_config_item(CONFIG, 'app', 'pagure_url').rstrip('/')
    pagure_api_url = '{0}/api/0'.format(pagure_url)
    pagure_repo_issue_url = '{0}/dist-git-requests/issue/{1}'.format(
        pagure_api_url, issue_id)

    issue_rv = requests_wrapper(
        pagure_repo_issue_url, timeout=60, service_name='Pagure')
    if issue_rv.status_code == 404:
        return None
    return get_request_json(issue_rv, 'getting an issue', 'error')


def get_issues():
    """
    List all the issues in ticket queue
    :return: a list of dictionaries containing issues
    """
    pagure_url = get_config_item(CONFIG, 'app', 'pagure_url').rstrip('/')
    pagure_api_url = '{0}/api/0'.format(pagure_url)
    pagure_repo_issues_url = '{0}/dist-git-requests/issues?{1}'.format(
        pagure_api_url, urlencode({'status': 'Open'}))

    issues_rv = requests_wrapper(
        pagure_repo_issues_url, timeout=60, service_name='Pagure')
    return get_request_json(
        issues_rv, 'getting open issues', 'error')['issues']


def get_pagure_issue_url(issue_id):
    """
    Get the URL to the issue in the UI
    :param issue_id: a string of the issue ID
    :return: a string of the URL to the issue in the UI
    """
    url = get_config_item(CONFIG, 'app', 'pagure_url').rstrip('/')
    return '{0}/dist-git-requests/issue/{1}'.format(url, issue_id)


def get_project_git_url(namespace, repo, url_type='ssh'):
    """
    Get the Git URL of a project
    :param namespace: a string representing the namespace to create the project
    in
    :param repo: a string of the project/repo name
    :param url_type: a string of the type of Git URL (e.g. ssh, git, etc.)
    :return: a dictionary of Git URLs
    """
    pagure_url = get_config_item(
        CONFIG, 'app', 'pagure_dist_git_url').rstrip('/')
    pagure_git_url_api_url = '{0}/api/0/{1}/{2}/git/urls'.format(
        pagure_url, namespace, repo)

    rv = requests_wrapper(
        pagure_git_url_api_url, timeout=60, service_name='Pagure')
    rv_json = get_request_json(rv, 'getting a project\'s git url', 'error')
    return rv_json['urls'].get(url_type)


def new_issue(title, body):
    """
    Posts a new Pagure issue
    :param title: string representing the issue's title
    :param body: string representing the issue's body
    :return: string of the URL to the created issue in the UI
    """
    pagure_url = get_config_item(CONFIG, 'app', 'pagure_url').rstrip('/')
    pagure_api_url = '{0}/api/0'.format(pagure_url)
    pagure_new_issue_url = '{0}/dist-git-requests/new_issue'.format(
        pagure_api_url)

    headers = get_pagure_auth_header()
    payload = {
        'title': title,
        'issue_content': body
    }

    rv = requests_wrapper(pagure_new_issue_url, data=payload, headers=headers,
                          timeout=60, http_verb='post', service_name='Pagure')
    rv_json = get_request_json(rv, 'creating a new issue', 'error')

    return '{0}/dist-git-requests/issue/{1}'.format(
        pagure_url, rv_json['issue']['id'])


def new_project(namespace, repo, description, upstreamurl):
    """
    Create a new Pagure project
    :param namespace: a string representing the namespace to create the project
    in
    :param repo: a string of the project/repo name
    :param description: a string of the description of the project
    :param upstreamurl: a string of the URL of the upstream project
    :return: None
    """
    pagure_url = get_config_item(CONFIG, 'app', 'pagure_dist_git_url')
    pagure_new_project_url = \
        '{0}/api/0/new'.format(pagure_url.rstrip('/'))

    headers = get_pagure_auth_header('admin', token_type='global')
    payload = {
        'namespace': namespace,
        'name': repo,
        'description': description or 'The {0} package'.format(repo),
        'url': upstreamurl or '',
        'create_readme': True,
        'wait': True
    }

    rv = requests_wrapper(
        pagure_new_project_url, data=payload, headers=headers, timeout=90,
        http_verb='post', service_name='Pagure')
    # We won't actually use the returned output from this function call but
    # it does error checking for us
    get_request_json(rv, 'creating a new project in Pagure')
    return None


def add_comment_to_issue(issue_id, comment):
    url = get_config_item(CONFIG, 'app', 'pagure_url').rstrip('/')
    issue_api_url = '{0}/api/0/dist-git-requests/issue/{1}'.format(
        url, issue_id)
    issue_comment_api_url = '{0}/comment'.format(issue_api_url)
    headers = get_pagure_auth_header('admin')
    comment_payload = {'comment': comment}
    rv = requests_wrapper(
        issue_comment_api_url, data=comment_payload, headers=headers,
        timeout=60, http_verb='post', service_name='Pagure')
    # We won't use the JSON but it gives us good error checking
    get_request_json(rv, 'adding a comment to an issue', 'error')


def close_issue(issue_id, comment=None, close_status=None):
    """
    Close a Pagure issue
    :param issue_id: the id of the ticket to close
    :param comment: the comment to leave in the issue before closing
    :param close_status: the close status of the issue. These have to be
    configured in the project.
    :return: None
    """
    url = get_config_item(CONFIG, 'app', 'pagure_url').rstrip('/')
    issue_api_url = '{0}/api/0/dist-git-requests/issue/{1}'.format(
        url, issue_id)
    issue_status_api_url = '{0}/status'.format(issue_api_url)
    headers = get_pagure_auth_header('admin')

    if comment is not None:
        add_comment_to_issue(issue_id, comment)

    status_payload = {'status': 'Closed'}
    if close_status is not None:
        status_payload['close_status'] = close_status
    rv = requests_wrapper(
        issue_status_api_url, data=status_payload, headers=headers,
        timeout=60, http_verb='post', service_name='Pagure')
    # We won't use the JSON but it gives us good error checking
    get_request_json(rv, 'changing the status of an issue', 'error')


def change_project_main_admin(namespace, repo, new_main_admin):
    """
    Change the main admin/owner of a project
    :param namespace: a string representing the namespace to create the project
    in
    :param repo: a string of the project/repo name
    :param new_main_admin: a string of the username of the new main admin
    :return: None
    """
    pagure_url = get_config_item(CONFIG, 'app', 'pagure_dist_git_url')
    pagure_modify_project_url = \
        '{0}/api/0/{1}/{2}'.format(pagure_url.rstrip('/'), namespace, repo)

    headers = get_pagure_auth_header('admin', token_type='global')
    payload = {
        'main_admin': new_main_admin
    }
    rv = requests_wrapper(
        pagure_modify_project_url, data=payload, headers=headers,
        timeout=60, http_verb='patch', service_name='Pagure')
    get_request_json(rv, 'setting the owner of a project in Pagure')
    return None
